from os.path import join
try:
    from helper_functions import  *
except ModuleNotFoundError:
    from pypet2bids.helper_functions import *


parent_dir = pathlib.Path(__file__).parent.resolve()
project_dir = parent_dir.parent
metadata_dir = join(project_dir, 'metadata')
pet_metadata_json = join(metadata_dir, 'PET_metadata.json')
permalink_pet_metadata_json = "https://github.com/openneuropet/PET2BIDS/blob/76d95cf65fa8a14f55a4405df3fdec705e2147cf/metadata/PET_metadata.json"

"""

:format:
:param:
:return:

Anthony Galassi
-----------------------------
Copyright Open NeuroPET team
"""


def translate_metadata(metadata_dataframe, image_path=NotImplemented):
    """
    This method exists as an example/template for an individual to customize and use on their own PET metadata contained
    within a spreadsheet. It is important to note that everything but the method name `translate_metadata` and the
    dictionary structure it returns can be modified to extract, transform, and load data from a user's spreadsheet.
    That is to say keep this method named `translate_metadata` and return a dictionary of the form:
    {
        'nifti_json: {},
        'blood_json: {},
        'blood_tsv: <pandas.DataFrame> or a <str>,
    }

    :param metadata_dataframe:
    :param image_path:
    :return:
    """
    nifti_json = {
        'Manufacturer': '',
        'ManufacturersModelName': '',
        'Units': '',
        'TracerName': '',
        'TracerRadionuclide': '',
        'InjectedRadioactivity': 0,
        'InjectedRadioactivityUnits': '',
        'InjectedMass': 0,
        'InjectedMassUnits': '',
        'SpecificRadioactivity': 0,
        'SpecificRadioactivityUnits': '',
        'ModeOfAdministration': '',
        'TimeZero': 0,
        'ScanStart': 0,
        'InjectionStart': 0,
        'FrameTimesStart': [],
        'FrameDuration': [],
        'AcquisitionMode': '',
        'ImageDecayCorrected': '',
        'ImageDecayCorrectionTime': 0,
        'ReconMethodName': '',
        'ReconMethodParameterLabels': [],
        'ReconMethodParameterUnits': [],
        'ReconMethodParameterValues': [],
        'ReconFilterType': '',
        'ReconFilterSize': 0,
        'AttenuationCorrection': '',
        'InstitutionName': '',
        'InstitutionalDepartmentName': ''
    }

    for key in nifti_json.keys():
        try:
            nifti_json[key] = flatten_series(metadata_dataframe[key])
        except KeyError:
            warnings.warn(f"{key} not found in metadata extracted from spreadsheet")

    blood_json = {
        "PlasmaAvail": True,
        "WholeBloodAvail": True,
        "MetaboliteAvail": False,
        "DispersionCorrected": False,
        "time": {
            "Description": "Time in relation to time zero defined in _pet.json",
            "Units": "s"
        },
        "plasma_radioactivity": {
            "Description": "Radioactivity in plasma samples, measured by eye balling it.",
            "Units": "kBq/mL"
        },
        "whole_blood_radioactivity": {
            "Description": "Radioactivity in whole blood samples, measured by divining rod.",
            "Units": "kBq/mL"
        }
    }

    blood_tsv = {
        "time": [],
        "plasma_radioactivity": [],
        "whole_blood_radioactivity": []
    }

    for key in blood_tsv.keys():
        try:
            blood_tsv[key] = flatten_series(metadata_dataframe[key])
        except KeyError:
            warnings.warn(f"{key} not found in metadata extracted from spreadsheet")

    # now transform the key value pairs in blood_tsv into a pandas dataframe object.
    blood_tsv = pandas.DataFrame.from_dict(blood_tsv)

    return {'nifti_json': nifti_json, 'blood_json': blood_json, 'blood_tsv': blood_tsv}



from json_maj.main import JsonMAJ

try:
    from pypet2bids.helper_functions import *
except ModuleNotFoundError:
    from helper_functions import *


def read_multi_subject_spreadsheets(
        general_metadata_spreadsheet: pathlib.Path,
        multiple_subject_spreadsheet: pathlib.Path,
        **kwargs) -> dict:
    """
    Reads in two spreadsheets as formatted in PET2BIDS/spreadsheet_conversion/many_subject_sheets,
    generic (scanner or subject independent data is supplied via the first argument and subject
    specific data is supplied via the second argument.

    :param general_metadata_spreadsheet: path to a metadata spreadsheet containing bids fields as columns
        with values below
    :type general_metadata_spreadsheet: file path
    :param multiple_subject_spreadsheet: path to multi subject spreadsheet containing a subject id, participant id,
        subject, or participant column consisting of paths to subject folders/files.
    :type multiple_subject_spreadsheet: file path
    :param kwargs: additional key pair arguments to pass on, these get applied generally
        just like the first spreadsheet. e.g. TimeZero=12:12:12, SpecificRadioactivity=3
    :type kwargs: string or dict
    :return: dictionary of subject data extracted from each spreadsheet along with any additional
        kwargs supplied
    :rtype: dict
    
    Anthony Galassi
    Copyright Open NeuroPET team

    """

    required_fields = load_pet_bids_requirements_json()

    if general_metadata_spreadsheet.is_file() and multiple_subject_spreadsheet.is_file():
        general_metadata = single_spreadsheet_reader(general_metadata_spreadsheet)
        multiple_subject_metadata = open_meta_data(multiple_subject_spreadsheet)
        multiple_subject_metadata

        column_set = set(list(general_metadata.keys()) + list(multiple_subject_metadata.columns))

        column_set = list(column_set)

        # assert required columns are in input spreadsheets
        for k in required_fields.keys():
            for field in required_fields[k]:
                field_exists = field in column_set
                if k == 'mandatory' and not field_exists:
                    logging.warning(f"Input spreadsheet(s) {general_metadata_spreadsheet} and "
                                    f"{multiple_subject_spreadsheet} are missing required column {field}")
                elif k == 'recommended' and not field_exists:
                    logging.info(f"Input spreadsheet(s) {general_metadata_spreadsheet} and "
                                 f"{multiple_subject_spreadsheet} are missing recommended column {field}")
                elif k == 'optional':
                    logging.info(f"Input spreadsheet(s) {general_metadata_spreadsheet} and "
                                 f"{multiple_subject_spreadsheet} are missing optional column {field}")

        # check to see if there's a subject column in the multi subject data
        accepted_column_names = ['participant_id', 'participant', 'subject', 'subject_id']
        found_column_names = []
        for acceptable in accepted_column_names:
            if acceptable in multiple_subject_metadata.columns:
                found_column_names.append(acceptable)

        if len(found_column_names) != 1:
            error_message = f"multi-subject_spreadsheet: {multiple_subject_spreadsheet} must contain only one column " \
                            f"of the following names: "
            for name in accepted_column_names:
                error_message += name + " "
            error_message += f"\nContains these instead {found_column_names}."
            raise Exception(error_message)
        else:
            subject_column = found_column_names[0]

        # collect all subject id's
        subject_metadata = {}
        for subject in multiple_subject_metadata.get(subject_column, None):
            subject_row = get_coordinates_containing(subject, multiple_subject_metadata, single=True)[0]
            subject_id = collect_bids_part('sub', subject)
            session_id = collect_bids_part('ses', subject)
            if subject_id:
                subject_metadata[subject_id] = general_metadata
                if session_id:
                    subject_metadata[subject_id]['session_id'] = session_id
                if kwargs:
                    subject_metadata[subject_id].update(**kwargs)

                subject_data_from_row = transform_row_to_dict(subject_row, multiple_subject_metadata)
                for k, v in subject_data_from_row.items():
                    if v and k != subject_column and v is not numpy.nan:
                        subject_metadata[subject_id][k] = v

        return subject_metadata

    else:
        missing_files = \
            [missing for missing in (general_metadata_spreadsheet, multiple_subject_spreadsheet)
             if not missing.is_file()]
        error_message = f"Missing "
        for m in missing_files:
            error_message += f'{m=}'.split('=')[0]
            error_message += f' {m}'
        raise Exception(error_message)


def write_multi_subject_spreadsheets(subjects: dict, output_path: typing.Union[str, pathlib.Path],
                                     create_bids_tree: bool = False) -> None:
    """
    Writes out a dictionary of subjects to a series of json files, if files exist updates
    them with new values obtained from spreadsheets.

    :param subjects: subject dictionary with subject id as primary keys and all bids fields as values
    :type subjects: dict
    :param output_path: path to write out files to, very much required
    :type output_path: str or pathlib.Path object
    :param create_bids_tree: boolean flag to create a bids tree, function will do it's best
        to create subject folders and session folders along w/ pet modality folder if this option is
        specified. Works on existing bids trees so long as session and subject id can be parsed from
        multi subject input sheet.
    :type create_bids_tree: bool
    :return: None
    :rtype: None
    """
    if create_bids_tree:
        for subject, fields in subjects.items():
            json_out_path = os.path.join(output_path, f"{subject}")
            if fields.get('session_id', None):
                json_out_file_name = subject + "_" + f"{fields.get('session_id')}_pet.json"
                json_out_path = os.path.join(json_out_path, fields.get('session_id'), 'pet')
            else:
                json_out_file_name = subject + "_pet.json"
                json_out_path = os.path.join(json_out_path, 'pet')
            pathlib.Path(json_out_path).mkdir(parents=True, exist_ok=True)
            json_out_file_name = os.path.join(json_out_path, json_out_file_name)
            json_out = JsonMAJ(json_path=json_out_file_name, update_values=fields).update()
    else:
        for subject, fields in subjects.items():
            json_out_path = os.path.join(output_path, f"{subject}")
            if fields.get('session_id', None):
                json_out_path += f"{fields.get('session_id')}_pet.json"
            else:
                json_out_path += '_pet.json'

            json_out = JsonMAJ(json_path=json_out_path, update_values=fields).update()


def cli():
    """
    Instantiates a command line interface with which allows reading and writing of
    2 spreadsheets following the format specified in PET2BIDS/spreadsheet_conversion/many_subjects_sheet
    :return: a dictionary version of the subject sidecar json's that get written out
    :rtype: dict
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--general-spreadsheet", '-g', type=pathlib.Path,
                        help="Path to a spreadsheet with data applicable to mulitiple subjects")
    parser.add_argument("--many-subjects-spreadsheet", '-m', type=pathlib.Path,
                        help="Path to spreadsheet containing multiple subjects")
    parser.add_argument("--output-path", "-o", type=pathlib.Path)
    parser.add_argument("--bids-tree", "-b", action='store_true')
    args = parser.parse_args()
    subjects = read_multi_subject_spreadsheets(
        general_metadata_spreadsheet=args.general_spreadsheet,
        multiple_subject_spreadsheet=args.many_subjects_spreadsheet)
    if args.output_path:
        output_path = args.output_path
    else:
        output_path = os.getcwd()

    write_multi_subject_spreadsheets(output_path=output_path, subjects=subjects, create_bids_tree=args.bids_tree)

    return subjects


if __name__ == '__main__':
    x = cli()



"""
Contains class ecat2ii, used to write ecat information out to a nifti file, depends on
`nibabel <https://nipy.org/nibabel/>`_ and  :func:`pypet2bids.read_ecat.read_ecat`.

| *Authors: Anthony Galassi*
| *Copyright OpenNeuroPET team*
"""
import datetime
import nibabel
import numpy
import pathlib
from pypet2bids.read_ecat import read_ecat
import os
import pickle
import logging

try:
    import helper_functions
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions


#logger = helper_functions.logger('pypet2bids')

logger = logging.getLogger('pypet2bids')

def ecat2nii(ecat_main_header=None,
             ecat_subheaders=None,
             ecat_pixel_data=None,
             ecat_file=None,
             nifti_file: str = '',
             sif_out=False,
             affine=None,
             save_binary=False,
             **kwargs):
    """
    Converts an ECAT file into a nifti and a sidecar json, used in conjunction with
    :func:`pypet2bids.read_ecat.read_ecat`
    
    :param ecat_main_header: the main header of an ECAT file
    :param ecat_subheaders: the subheaders for each frame of the ECAT file
    :param ecat_pixel_data: the imaging/pixel data from the ECAT file
    :param ecat_file: the path to the ECAT file, required and used to create .nii and .json output files
    :param nifti_file: the desired output path of the nifti file
    :param sif_out: outputs a .sif file containing the images pixel data
    :param affine: a user supplied affine, this is gathered from the ECAT if not supplied.
    :param save_binary: dumps a pickled ECAT object, you probably shouldn't be using this.
    :param kwargs: additional key value pairs that one wishes to add to the sidecar json accompanying a converted ECAT
        nii image
    :return: a nibabel nifti object if one wishes to muddle with the object in python and not in a .nii file

    """

    # if a nifti file/path is not included write a nifti next to the ecat file
    if not nifti_file:
        nifti_file = os.path.splitext(ecat_file)[0] + ".nii"
    else:
        nifti_file = nifti_file

    if not pathlib.Path(nifti_file).parent.exists():
        pathlib.Path(nifti_file).parent.mkdir(parents=True, exist_ok=True)

    # collect the output folder from the nifti path will use for .sif files
    output_folder = pathlib.Path(nifti_file).parent
    nifti_file_w_out_extension = os.path.splitext(str(pathlib.Path(nifti_file).name))[0]

    # if already read nifti file skip re-reading
    if ecat_main_header is None and ecat_subheaders is None and ecat_pixel_data is None and ecat_file:
        # collect ecat_file
        main_header, sub_headers, data = read_ecat(ecat_file=ecat_file)
    elif ecat_file is None and type(ecat_main_header) is dict and type(ecat_subheaders) is list and type(
            ecat_pixel_data) is numpy.ndarray:
        main_header, sub_headers, data = ecat_main_header, ecat_subheaders, ecat_pixel_data
    else:
        raise Exception("Must pass in filepath for ECAT file or "
                        "(ecat_main_header, ecat_subheaders, and ecat_pixel data "
                        f"got ecat_file={ecat_file}, type(ecat_main_header)={type(ecat_main_header)}, "
                        f"type(ecat_subheaders)={type(ecat_subheaders)}, "
                        f"type(ecat_pixel_data)={type(ecat_pixel_data)} instead.")

    # check for TimeZero supplied via kwargs
    if kwargs.get('TimeZero', None):
        TimeZero = kwargs['TimeZero']
    else:
        logger.warn("Metadata TimeZero is missing -- set to ScanStart or empty to use the scanning time as "
              "injection time")

    # get image shape
    img_shape = data.shape
    shape_from_headers = (sub_headers[0]['X_DIMENSION'],
                          sub_headers[0]['Y_DIMENSION'],
                          sub_headers[0]['Z_DIMENSION'],
                          main_header['NUM_FRAMES'])

    # make sure number of data elements matches frame number
    single_frame = False
    if img_shape[3] == 1 and img_shape[0:2] == shape_from_headers[0:2]:
        single_frame = True
    if img_shape != shape_from_headers and not single_frame:
        raise Exception(
            f"Mis-match between expected X,Y,Z, and Num. Frames dimensions ({shape_from_headers} obtained from headers"
            f"and shape of imaging data ({img_shape}")

    # format data into acceptable shape for nibabel, by first creating empty matrix
    img_temp = numpy.zeros(shape=(sub_headers[0]['X_DIMENSION'],
                                  sub_headers[0]['Y_DIMENSION'],
                                  sub_headers[0]['Z_DIMENSION'],
                                  main_header['NUM_FRAMES']),
                           dtype=numpy.dtype('>f4'))

    # collect timing information
    start, delta = [], []

    # collect prompts and randoms
    prompts, randoms = [], []

    # load frame data into img temp
    for index in reversed(range(img_shape[3])):  # Don't throw stones working from existing matlab code
        print(f"Loading frame {index + 1}")
        # save out our slice of data before flip to a text file to compare w/ matlab data
        img_temp[:, :, :, index] = numpy.flip(numpy.flip(numpy.flip(
            data[:, :, :, index].astype(numpy.dtype('>f4')) * sub_headers[index]['SCALE_FACTOR'], 1), 2), 0)
        start.append(sub_headers[index]['FRAME_START_TIME'] * 60)  # scale to per minute
        delta.append(sub_headers[index]['FRAME_DURATION'] * 60)  # scale to per minute

        if main_header.get('SW_VERSION', 0) >= 73:
            # scale both to per minute
            prompts.append(sub_headers[index]['PROMPT_RATE'] * sub_headers[index]['FRAME_DURATION'] * 60)
            randoms.append(sub_headers[index]['RANDOM_RATE'] * sub_headers[index]['FRAME_DURATION'] * 60)
        else:
            # this field is not available in ecat 7.2
            prompts.append(0)
            randoms.append(0)

    final_image = img_temp * main_header['ECAT_CALIBRATION_FACTOR']

    qoffset_x = -1 * (
        ((sub_headers[0]['X_DIMENSION'] * sub_headers[0]['X_PIXEL_SIZE'] * 10 / 2) - sub_headers[0][
            'X_PIXEL_SIZE'] * 5))

    qoffset_y = -1 * (
        ((sub_headers[0]['Y_DIMENSION'] * sub_headers[0]['Y_PIXEL_SIZE'] * 10 / 2) - sub_headers[0][
            'Y_PIXEL_SIZE'] * 5))

    qoffset_z = -1 * (
        ((sub_headers[0]['Z_DIMENSION'] * sub_headers[0]['Z_PIXEL_SIZE'] * 10 / 2) - sub_headers[0][
            'Z_PIXEL_SIZE'] * 5))

    # build affine if it's not included in function call
    if not affine:
        t = numpy.identity(4)
        t[0, 0] = sub_headers[0]['X_PIXEL_SIZE'] * 10
        t[1, 1] = sub_headers[0]['Y_PIXEL_SIZE'] * 10
        t[2, 2] = sub_headers[0]['Z_PIXEL_SIZE'] * 10

        t[3, 0] = qoffset_x
        t[3, 1] = qoffset_y
        t[3, 2] = qoffset_z

        # note this affine is the transform of of a nibabel ecat object's affine
        affine = t

    img_nii = nibabel.Nifti1Image(final_image, affine=affine)

    # populating nifti header
    if img_nii.header['sizeof_hdr'] != 348:
        img_nii.header['sizeof_hdr'] = 348
    # img_nii.header['dim_info'] is populated on object creation
    # img_nii.header['dim']  is populated on object creation
    img_nii.header['intent_p1'] = 0
    img_nii.header['intent_p2'] = 0
    img_nii.header['intent_p3'] = 0
    # img_nii.header['datatype'] # created on invocation seems to be 16 or int16
    # img_nii.header['bitpix'] # also automatically created and inferred 32 as of testing w/ cimbi dataset
    # img_nii.header['slice_type'] # defaults to 0
    # img_nii.header['pixdim'] # appears as 1d array of length 8 we rescale this
    img_nii.header['pixdim'] = numpy.array(
        [1,
         sub_headers[0]['X_PIXEL_SIZE'] * 10,
         sub_headers[0]['Y_PIXEL_SIZE'] * 10,
         sub_headers[0]['Z_PIXEL_SIZE'] * 10,
         0,
         0,
         0,
         0])
    img_nii.header['vox_offset'] = 352

    # TODO img_nii.header['scl_slope'] # this is a NaN array by default but apparently it should be the dose calibration
    #  factor img_nii.header['scl_inter'] # defaults to NaN array
    img_nii.header['scl_slope'] = main_header['ECAT_CALIBRATION_FACTOR']
    img_nii.header['scl_inter'] = 0
    img_nii.header['slice_end'] = 0
    img_nii.header['slice_code'] = 0
    img_nii.header['xyzt_units'] = 10
    img_nii.header['cal_max'] = final_image.min()
    img_nii.header['cal_min'] = final_image.max()
    img_nii.header['slice_duration'] = 0
    img_nii.header['toffset'] = 0
    img_nii.header['descrip'] = "OpenNeuroPET ecat2nii.py conversion"
    # img_nii.header['aux_file'] # ignoring as this is set to '' in matlab
    img_nii.header['qform_code'] = 0
    img_nii.header['sform_code'] = 1  # 0: Arbitrary coordinates;
    # 1: Scanner-based anatomical coordinates;
    # 2: Coordinates aligned to another file's, or to anatomical "truth" (co-registration);
    # 3: Coordinates aligned to Talairach-Tournoux Atlas; 4: MNI 152 normalized coordinates

    img_nii.header['quatern_b'] = 0
    img_nii.header['quatern_c'] = 0
    img_nii.header['quatern_d'] = 0
    # Please explain this
    img_nii.header['qoffset_x'] = qoffset_x
    img_nii.header['qoffset_y'] = qoffset_y
    img_nii.header['qoffset_z'] = qoffset_z
    img_nii.header['srow_x'] = numpy.array([sub_headers[0]['X_PIXEL_SIZE']*10, 0, 0, img_nii.header['qoffset_x']])
    img_nii.header['srow_y'] = numpy.array([0, sub_headers[0]['Y_PIXEL_SIZE']*10, 0, img_nii.header['qoffset_y']])
    img_nii.header['srow_z'] = numpy.array([0, 0, sub_headers[0]['Z_PIXEL_SIZE']*10, img_nii.header['qoffset_z']])

    img_nii.header['intent_name'] = ''
    img_nii.header['magic'] = 'n + 1 '

    # nifti header items to include
    img_nii.header.set_xyzt_units('mm', 'unknown')

    # save nifti
    nibabel.save(img_nii, nifti_file)

    # used for testing veracity of nibabel read and write.
    if save_binary:
        pickle.dump(img_nii, open(nifti_file + '.pickle', "wb"))

    # write out timing file
    if sif_out:
        with open(os.path.join(output_folder, nifti_file_w_out_extension + '.sif'), 'w') as sif_file:
            scantime = datetime.datetime.fromtimestamp(main_header['SCAN_START_TIME'])
            scantime = scantime.astimezone().isoformat()
            sif_file.write(f"{scantime} {len(start)} 4 1\n")
            for index in reversed(range(len(start))):
                start_i = round(start[index])
                start_i_plus_delta_i = start_i + round(delta[index])
                prompt = round(prompts[index])
                random = round(randoms[index])
                output_string = f"{start_i} {start_i_plus_delta_i} {prompt} {random}\n"
                sif_file.write(output_string)

    return img_nii
